#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define N 3

void makemat(double A[][N], double *b, char* fn)
{
  FILE *fp;
  int i, j;
  char s[N*10];
  fp = fopen(fn, "r");
  if(fp == NULL){
	perror("fopen:");
	exit(EXIT_FAILURE);
  }
  // matrix
  for(i=0;i<N; i++){
	char *ret, *tp;
	ret = fgets(s, N*10, fp);
	if(ret==NULL){
	  perror("fgets:");
	  exit(EXIT_FAILURE);
	}
	tp = strtok(s, " ");
	for(j=0; j<N; j++){
	  sscanf(tp, "%lf", &A[i][j]);
	  tp = strtok(NULL, " ");
	}
  }
  // vector
  {
	char *ret, *tp;
	ret = fgets(s, N*10, fp);
	if(ret==NULL){
	  perror("fgets:");
	  exit(EXIT_FAILURE);
	}
	tp = strtok(s, " ");
	for(j=0; j<N; j++){
	  sscanf(tp, "%lf", &b[j]);
	  tp = strtok(NULL, " ");
	}
  }
  fclose(fp);
}

// partial pivoting
// with data exchange
void forward_elimination_1(double A[][N], double *b)
{
  int i, j, k;
  double p;
  for(i=0; i<N; i++){

	// search
	double maxp = 0.0;
	int y = i;
	double tmp;
	for(j=i; j<N; j++){
	  if(maxp<fabs(A[j][i])){
		maxp = fabs(A[j][i]);
		y = j;
	  }
	}

	// data exchange
	{
	  for(j=i; j<N; j++){
		tmp = A[i][j];
		A[i][j] = A[y][j];
		A[y][j] = tmp;
	  }
	  tmp = b[i];
	  b[i] = b[y];
	  b[y] = tmp;
	}

	p = A[i][i];
	for(j=i; j<N; j++){
	  A[i][j] /= p;
	}
	b[i] /= p;

	for(j=i+1; j<N; j++){
	  p = A[j][i];
	  for(k=i; k<N; k++){
		A[j][k] -= A[i][k] * p;
	  }
	  b[j] -= b[i] * p;
	}
  }
}

void backward_substitution_1(double A[][N], double *b)
{
  int i, j;
  double p;
  for(i=N-1; i>=0; i--){
	p = 0.0;
	for(j=i+1; j<N; j++){
	  p += A[i][j] * b[j];
	}
	b[i] -= p;
  }
}

void printresult_1(double A[][N], double *b)
{
  int i, j;
  printf("A\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  printf(" %f", A[i][j]);
	}
	printf("\n");
  }

  printf("b(result)\n");
  for(i=0; i<N; i++){
	printf(" %f", b[i]);
  }
  printf("\n");

  printf("\n");
}

// partial pivoting
// without data exchange
void forward_elimination_2(double A[][N], double *b, int * index)
{
  int i, j, k;
  double p;

  for(i=0; i<N; i++){
	index[i] = i;
  }

  for(i=0; i<N; i++){

	// search
	double maxp = 0.0;
	int y = i;
	int tmp;
	for(j=i; j<N; j++){
	  if(maxp<fabs(A[index[j]][i])){
		maxp = fabs(A[index[j]][i]);
		y = j;
	  }
	}
	// make indirect index
	{
	  tmp = index[y];
	  index[y] = index[i];
	  index[i] = tmp;
	}

	p = A[index[i]][i];
	for(j=i; j<N; j++){
	  A[index[i]][j] /= p;
	}
	b[index[i]] /= p;

	for(j=i+1; j<N; j++){
	  p = A[index[j]][i];
	  for(k=i; k<N; k++){
		A[index[j]][k] -= A[index[i]][k] * p;
	  }
	  b[index[j]] -= b[index[i]] * p;
	}
  }
}

void backward_substitution_2(double A[][N], double *b, int * index)
{
  int i, j;
  double p;
  for(i=N-1; i>=0; i--){
	p = 0.0;
	for(j=i+1; j<N; j++){
	  p += A[index[i]][j] * b[index[j]];
	}
	b[index[i]] -= p;
  }
}

void printresult_2(double A[][N], double *b, int * index)
{
  int i, j;
  printf("A\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  printf(" %f", A[index[i]][j]);
	}
	printf("\n");
  }

  printf("b\n");
  for(i=0; i<N; i++){
	printf(" %f", b[i]);
  }
  printf("\n");

  printf("index\n");
  for(i=0; i<N; i++){
	printf(" %d", index[i]);
  }
  printf("\n");

  printf("result:\n");
  for(i=0; i<N; i++){
	printf(" %f", b[index[i]]);
  }
  printf("\n");

  printf("\n");
}


int main(int argc, char **argv){
  double A[N][N], b[N];
  int *index;
  if(argc!=2){
	printf("usage: %s datafile\n", argv[0]);
	return(EXIT_FAILURE);
  }

  printf("with data exchange\n");
  makemat(A, b, argv[1]);
  forward_elimination_1(A, b);
  backward_substitution_1(A, b);
  printresult_1(A, b);

  printf("without data exchange\n");
  makemat(A, b, argv[1]);
  index = (int*)malloc(sizeof(int)*N);
  forward_elimination_2(A, b, index);
  backward_substitution_2(A, b, index);
  printresult_2(A, b, index);
  free(index);

  return EXIT_SUCCESS;
}
