#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define N 3

void makemat(double A[][N], double *b, char* fn)
{
  FILE *fp;
  int i, j;
  char s[N*10];
  fp = fopen(fn, "r");
  if(fp == NULL){
	perror("fopen:");
	exit(EXIT_FAILURE);
  }
  // matrix
  for(i=0;i<N; i++){
	char *ret, *tp;
	ret = fgets(s, N*10, fp);
	if(ret==NULL){
	  perror("fgets:");
	  exit(EXIT_FAILURE);
	}
	tp = strtok(s, " ");
	for(j=0; j<N; j++){
	  sscanf(tp, "%lf", &A[i][j]);
	  tp = strtok(NULL, " ");
	}
  }
  // vector
  {
	char *ret, *tp;
	ret = fgets(s, N*10, fp);
	if(ret==NULL){
	  perror("fgets:");
	  exit(EXIT_FAILURE);
	}
	tp = strtok(s, " ");
	for(j=0; j<N; j++){
	  sscanf(tp, "%lf", &b[j]);
	  tp = strtok(NULL, " ");
	}
  }
  fclose(fp);
}

// full pivoting
// with data exchange
void forward_elimination_1(double A[][N], double *b, int *cind)
{
  int i, j, k;
  double p;

  for(i=0; i<N; i++){
	cind[i] = i;
  }

  for(i=0; i<N; i++){

	// search
	double maxp = 0.0;
	int y = i;
	int x = i;
	int ntmp;
	double dtmp;
	for(j=i; j<N; j++){
	  for(k=i; k<N; k++){
		if(maxp<fabs(A[j][k])){
		  maxp = fabs(A[j][k]);
		  y = j;
		  x = k;
		}
	  }
	}
	printf("pivot is A[%d][%d] %f\n", y, x, maxp);

	// data exchange
	for(j=0; j<N; j++){
	  dtmp = A[i][j];
	  A[i][j] = A[y][j];
	  A[y][j] = dtmp;
	}

	dtmp = b[i];
	b[i] = b[y];
	b[y] = dtmp;

	for(j=0; j<N; j++){
	  dtmp = A[j][i];
	  A[j][i] = A[j][x];
	  A[j][x] = dtmp;
	}

	ntmp = cind[i];
	cind[i] = cind[x];
	cind[x] = ntmp;

	p = A[i][i];
	for(j=i; j<N; j++){
	  A[i][j] /= p;
	}
	b[i] /= p;

	for(j=i+1; j<N; j++){
	  p = A[j][i];
	  for(k=i; k<N; k++){
		A[j][k] -= A[i][k] * p;
	  }
	  b[j] -= b[i] * p;
	}
  }
}

void backward_substitution_1(double A[][N], double *b)
{
  int i, j;
  double p;
  for(i=N-1; i>=0; i--){
	p = 0.0;
	for(j=i+1; j<N; j++){
	  p += A[i][j] * b[j];
	}
	b[i] -= p;
  }
}

void printresult_1(double A[][N], double *b, int *cind)
{
  int i, j;
  printf("A\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  printf(" %f", A[i][j]);
	}
	printf("\n");
  }

  printf("b\n");
  for(i=0; i<N; i++){
	printf(" %f", b[i]);
  }
  printf("\n");

  printf("cind\n");
  for(i=0; i<N; i++){
	printf(" %d", cind[i]);
  }
  printf("\n");

  printf("result:\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  if(cind[j]==i){
		printf(" %f", b[j]);
		break;
	  }
	}
  }
  printf("\n");

  printf("\n");
}

// full pivoting
// without data exchange
void forward_elimination_2(double A[][N], double *b, int *rind, int * cind)
{
  int i, j, k;
  double p;

  for(i=0; i<N; i++){
	rind[i] = i;
	cind[i] = i;
  }

  for(i=0; i<N; i++){

	// search
	double maxp = 0.0;
	int y = i;
	int x = i;
	int ntmp;
	for(j=i; j<N; j++){
	  for(k=i; k<N; k++){
		if(maxp<fabs(A[j][k])){
		  maxp = fabs(A[j][k]);
		  y = j;
		  x = k;
		}
	  }
	}
	printf("pivot is A[%d][%d] %f\n", y, x, maxp);

	// make indirect index
	ntmp = rind[y];
	rind[y] = rind[i];
	rind[i] = ntmp;

	ntmp = cind[x];
	cind[x] = cind[i];
	cind[i] = ntmp;

	p = A[rind[i]][cind[i]];
	for(j=i; j<N; j++){
	  A[rind[i]][cind[j]] /= p;
	}
	b[rind[i]] /= p;

	for(j=i+1; j<N; j++){
	  p = A[rind[j]][cind[i]];
	  for(k=i; k<N; k++){
		A[rind[j]][cind[k]] -= A[rind[i]][cind[k]] * p;
	  }
	  b[rind[j]] -= b[rind[i]] * p;
	}
  }
}

void backward_substitution_2(double A[][N], double *b, int *rind, int * cind)
{
  int i, j;
  double p;
  for(i=N-1; i>=0; i--){
	p = 0.0;
	for(j=i+1; j<N; j++){
	  p += A[rind[i]][cind[j]] * b[rind[j]];
	}
	b[rind[i]] -= p;
  }
}

void printresult_2(double A[][N], double *b, int *rind, int * cind)
{
  int i, j;
  printf("A\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  printf(" %f", A[rind[i]][cind[j]]);
	}
	printf("\n");
  }

  printf("b\n");
  for(i=0; i<N; i++){
	printf(" %f", b[i]);
  }
  printf("\n");

  printf("rind\n");
  for(i=0; i<N; i++){
	printf(" %d", rind[i]);
  }
  printf("\n");

  printf("cind\n");
  for(i=0; i<N; i++){
	printf(" %d", cind[i]);
  }
  printf("\n");

  printf("result:\n");
  for(i=0; i<N; i++){
	for(j=0; j<N; j++){
	  if(cind[j]==i){
		printf(" %f", b[rind[j]]);
		break;
	  }
	}
  }
  printf("\n");

  printf("\n");
}


int main(int argc, char **argv){
  double A[N][N], b[N];
  int *rind;
  int *cind;

  if(argc!=2){
	printf("usage: %s datafile\n", argv[0]);
	return(EXIT_FAILURE);
  }

  printf("with data exchange\n");
  makemat(A, b, argv[1]);
  cind = (int*)malloc(sizeof(int)*N);
  forward_elimination_1(A, b, cind);
  backward_substitution_1(A, b);
  printresult_1(A, b, cind);
  free(cind);

  printf("without data exchange\n");
  makemat(A, b, argv[1]);
  rind = (int*)malloc(sizeof(int)*N);
  cind = (int*)malloc(sizeof(int)*N);
  forward_elimination_2(A, b, rind, cind);
  backward_substitution_2(A, b, rind, cind);
  printresult_2(A, b, rind,cind);
  free(rind);
  free(cind);

  return EXIT_SUCCESS;
}
